@@ -4,7 +4,6 @@ Original code can be found here: https://bitbucket.org/mvdschaar/mlforhealthlabp
 """
 from tensorflow import function, GradientTape, sqrt, abs, reduce_mean, ones_like, zeros_like, convert_to_tensor,float32
 from tensorflow import data as tfdata
-from tensorflow import config as tfconfig
 from tensorflow import nn
 from tensorflow.keras import Model, Sequential, Input
 from tensorflow.keras.layers import GRU, LSTM, Dense
@@ -14,7 +13,7 @@ from tensorflow.keras.losses import BinaryCrossentropy, MeanSquaredError
 import numpy as np
 from tqdm import tqdm, trange
 
-from ydata_synthetic.synthesizers import gan
+from ydata_synthetic.synthesizers.gan import BaseModel
 
 def make_net(model, n_layers, hidden_units, output_units, net_type='GRU'):
     if net_type=='GRU':
@@ -34,7 +33,7 @@ def make_net(model, n_layers, hidden_units, output_units, net_type='GRU'):
     return model
 
 
-class TimeGAN(gan.Model):
+class TimeGAN(BaseModel):
     def __init__(self, model_parameters, hidden_dim, seq_len, n_seq, gamma):
         self.seq_len=seq_len
         self.n_seq=n_seq
@@ -43,11 +42,11 @@ class TimeGAN(gan.Model):
         super().__init__(model_parameters)
 
     def define_gan(self):
-        self.generator_aux=Generator(self.hidden_dim).build(input_shape=(self.seq_len, self.n_seq))
-        self.supervisor=Supervisor(self.hidden_dim).build(input_shape=(self.hidden_dim, self.hidden_dim))
-        self.discriminator=Discriminator(self.hidden_dim).build(input_shape=(self.seq_len, self.n_seq))
-        self.recovery = Recovery(self.hidden_dim, self.n_seq).build(input_shape=(self.hidden_dim, self.hidden_dim))
-        self.embedder = Embedder(self.hidden_dim).build(input_shape=(self.hidden_dim, self.n_seq))
+        self.generator_aux=Generator(self.hidden_dim).build()
+        self.supervisor=Supervisor(self.hidden_dim).build()
+        self.discriminator=Discriminator(self.hidden_dim).build()
+        self.recovery = Recovery(self.hidden_dim, self.n_seq).build()
+        self.embedder = Embedder(self.hidden_dim).build()
 
         X = Input(shape=[self.seq_len, self.n_seq], batch_size=self.batch_size, name='RealData')
         Z = Input(shape=[self.seq_len, self.n_seq], batch_size=self.batch_size, name='RandomNoise')
@@ -95,15 +94,6 @@ class TimeGAN(gan.Model):
                                          outputs=Y_real,
                                          name="RealDiscriminator")
 
-        # ----------------------------
-        # Init the optimizers
-        # ----------------------------
-        self.autoencoder_opt = Adam(learning_rate=self.lr)
-        self.supervisor_opt = Adam(learning_rate=self.lr)
-        self.generator_opt = Adam(learning_rate=self.lr)
-        self.discriminator_opt = Adam(learning_rate=self.lr)
-        self.embedding_opt = Adam(learning_rate=self.lr)
-
         # ----------------------------
         # Define the loss functions
         # ----------------------------
@@ -112,7 +102,7 @@ class TimeGAN(gan.Model):
 
 
     @function
-    def train_autoencoder(self, x):
+    def train_autoencoder(self, x, opt):
         with GradientTape() as tape:
             x_tilde = self.autoencoder(x)
             embedding_loss_t0 = self._mse(x, x_tilde)
@@ -120,42 +110,47 @@ class TimeGAN(gan.Model):
 
         var_list = self.embedder.trainable_variables + self.recovery.trainable_variables
         gradients = tape.gradient(e_loss_0, var_list)
-        self.autoencoder_opt.apply_gradients(zip(gradients, var_list))
+        opt.apply_gradients(zip(gradients, var_list))
         return sqrt(embedding_loss_t0)
 
     @function
-    def train_supervisor(self, x):
+    def train_supervisor(self, x, opt):
         with GradientTape() as tape:
             h = self.embedder(x)
             h_hat_supervised = self.supervisor(h)
-            g_loss_s = self._mse(h[:, 1:, :], h_hat_supervised[:, 1:, :])
+            generator_loss_supervised = self._mse(h[:, 1:, :], h_hat_supervised[:, :-1, :])
 
         var_list = self.supervisor.trainable_variables + self.generator.trainable_variables
-        gradients = tape.gradient(g_loss_s, var_list)
-        self.supervisor_opt.apply_gradients(zip(gradients, var_list))
-        return g_loss_s
+        gradients = tape.gradient(generator_loss_supervised, var_list)
+        apply_grads = [(grad, var) for (grad, var) in zip(gradients, var_list) if grad is not None]
+        opt.apply_gradients(apply_grads)
+        return generator_loss_supervised
 
     @function
-    def train_embedder(self,x):
+    def train_embedder(self,x, opt):
         with GradientTape() as tape:
+            # Supervised Loss
             h = self.embedder(x)
             h_hat_supervised = self.supervisor(h)
-            generator_loss_supervised = self._mse(h[:, 1:, :], h_hat_supervised[:, 1:, :])
+            generator_loss_supervised = self._mse(h[:, 1:, :], h_hat_supervised[:, :-1, :])
 
+            # Reconstruction Loss
             x_tilde = self.autoencoder(x)
             embedding_loss_t0 = self._mse(x, x_tilde)
             e_loss = 10 * sqrt(embedding_loss_t0) + 0.1 * generator_loss_supervised
 
         var_list = self.embedder.trainable_variables + self.recovery.trainable_variables
         gradients = tape.gradient(e_loss, var_list)
-        self.embedding_opt.apply_gradients(zip(gradients, var_list))
+        opt.apply_gradients(zip(gradients, var_list))
         return sqrt(embedding_loss_t0)
 
     def discriminator_loss(self, x, z):
+        # Loss on false negatives
         y_real = self.discriminator_model(x)
         discriminator_loss_real = self._bce(y_true=ones_like(y_real),
                                             y_pred=y_real)
 
+        # Loss on false positives
         y_fake = self.adversarial_supervised(z)
         discriminator_loss_fake = self._bce(y_true=zeros_like(y_fake),
                                             y_pred=y_fake)
@@ -176,7 +171,7 @@ class TimeGAN(gan.Model):
         return g_loss_mean + g_loss_var
 
     @function
-    def train_generator(self, x, z):
+    def train_generator(self, x, z, opt):
         with GradientTape() as tape:
             y_fake = self.adversarial_supervised(z)
             generator_loss_unsupervised = self._bce(y_true=ones_like(y_fake),
@@ -187,7 +182,7 @@ class TimeGAN(gan.Model):
                                                       y_pred=y_fake_e)
             h = self.embedder(x)
             h_hat_supervised = self.supervisor(h)
-            generator_loss_supervised = self._mse(h[:, 1:, :], h_hat_supervised[:, 1:, :])
+            generator_loss_supervised = self._mse(h[:, 1:, :], h_hat_supervised[:, :-1, :])
 
             x_hat = self.generator(z)
             generator_moment_loss = self.calc_generator_moments_loss(x, x_hat)
@@ -199,17 +194,17 @@ class TimeGAN(gan.Model):
 
         var_list = self.generator_aux.trainable_variables + self.supervisor.trainable_variables
         gradients = tape.gradient(generator_loss, var_list)
-        self.generator_opt.apply_gradients(zip(gradients, var_list))
+        opt.apply_gradients(zip(gradients, var_list))
         return generator_loss_unsupervised, generator_loss_supervised, generator_moment_loss
 
     @function
-    def train_discriminator(self, x, z):
+    def train_discriminator(self, x, z, opt):
         with GradientTape() as tape:
             discriminator_loss = self.discriminator_loss(x, z)
 
         var_list = self.discriminator.trainable_variables
         gradients = tape.gradient(discriminator_loss, var_list)
-        self.discriminator_opt.apply_gradients(zip(gradients, var_list))
+        opt.apply_gradients(zip(gradients, var_list))
         return discriminator_loss
 
     def get_batch_data(self, data, n_windows):
@@ -229,16 +224,22 @@ class TimeGAN(gan.Model):
 
     def train(self, data, train_steps):
         ## Embedding network training
+        autoencoder_opt = Adam(learning_rate=self.g_lr)
         for _ in tqdm(range(train_steps), desc='Emddeding network training'):
             X_ = next(self.get_batch_data(data, n_windows=len(data)))
-            step_e_loss_t0 = self.train_autoencoder(X_)
+            step_e_loss_t0 = self.train_autoencoder(X_, autoencoder_opt)
 
         ## Supervised Network training
+        supervisor_opt = Adam(learning_rate=self.g_lr)
         for _ in tqdm(range(train_steps), desc='Supervised network training'):
             X_ = next(self.get_batch_data(data, n_windows=len(data)))
-            step_g_loss_s = self.train_supervisor(X_)
+            step_g_loss_s = self.train_supervisor(X_, supervisor_opt)
 
         ## Joint training
+        generator_opt = Adam(learning_rate=self.g_lr)
+        embedder_opt = Adam(learning_rate=self.g_lr)
+        discriminator_opt = Adam(learning_rate=self.d_lr)
+
         step_g_loss_u = step_g_loss_s = step_g_loss_v = step_e_loss_t0 = step_d_loss = 0
         for _ in tqdm(range(train_steps), desc='Joint networks training'):
 
@@ -250,18 +251,18 @@ class TimeGAN(gan.Model):
                 # --------------------------
                 # Train the generator
                 # --------------------------
-                step_g_loss_u, step_g_loss_s, step_g_loss_v = self.train_generator(X_, Z_)
+                step_g_loss_u, step_g_loss_s, step_g_loss_v = self.train_generator(X_, Z_, generator_opt)
 
                 # --------------------------
                 # Train the embedder
                 # --------------------------
-                step_e_loss_t0 = self.train_embedder(X_)
+                step_e_loss_t0 = self.train_embedder(X_, embedder_opt)
 
             X_ = next(self.get_batch_data(data, n_windows=len(data)))
             Z_ = next(self.get_batch_noise())
             step_d_loss = self.discriminator_loss(X_, Z_)
             if step_d_loss > 0.15:
-                step_d_loss = self.train_discriminator(X_, Z_)
+                step_d_loss = self.train_discriminator(X_, Z_, discriminator_opt)
 
     def sample(self, n_samples):
         steps = n_samples // self.batch_size + 1
@@ -273,16 +274,13 @@ class TimeGAN(gan.Model):
         return np.array(np.vstack(data))
 
 
-
-
 class Generator(Model):
     def __init__(self, hidden_dim, net_type='GRU'):
         self.hidden_dim = hidden_dim
         self.net_type = net_type
 
-    def build(self, input_shape):
+    def build(self):
         model = Sequential(name='Generator')
-        model.add(Input(shape=input_shape))
         model = make_net(model,
                          n_layers=3,
                          hidden_units=self.hidden_dim,
@@ -295,7 +293,7 @@ class Discriminator(Model):
         self.hidden_dim = hidden_dim
         self.net_type=net_type
 
-    def build(self, input_shape):
+    def build(self):
         model = Sequential(name='Discriminator')
         model = make_net(model,
                          n_layers=3,
@@ -310,9 +308,8 @@ class Recovery(Model):
         self.n_seq=n_seq
         return
 
-    def build(self, input_shape):
+    def build(self):
         recovery = Sequential(name='Recovery')
-        recovery.add(Input(shape=input_shape, name='EmbeddedData'))
         recovery = make_net(recovery,
                             n_layers=3,
                             hidden_units=self.hidden_dim,
@@ -325,9 +322,8 @@ class Embedder(Model):
         self.hidden_dim=hidden_dim
         return
 
-    def build(self, input_shape):
+    def build(self):
         embedder = Sequential(name='Embedder')
-        embedder.add(Input(shape=input_shape, name='Data'))
         embedder = make_net(embedder,
                             n_layers=3,
                             hidden_units=self.hidden_dim,
@@ -338,9 +334,8 @@ class Supervisor(Model):
     def __init__(self, hidden_dim):
         self.hidden_dim=hidden_dim
 
-    def build(self, input_shape):
+    def build(self):
         model = Sequential(name='Supervisor')
-        model.add(Input(shape=input_shape))
         model = make_net(model,
                          n_layers=2,
                          hidden_units=self.hidden_dim,
